/*
INCLUDE -- select file entries from a list
    Copyright (C) 2003,2004  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

    Compile with Visual Studio 6 or later. cl include.c
    Compile with GCC 2.95.2 or later	   gcc -oinclude  (windows and linux)
    If compiling on MAC SYSTEM X (and possibly other BSD based os's), you will
    need to modify the include for malloc.h to sys/malloc.h

*/



#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h> 
#ifdef WIN32
#include <process.h> 
#else
#define stricmp strcasecmp
#define strnicmp strncasecmp
#endif

#define tolower(x) x

char s[32767];

char *stristr(char *s,char *t) {
	size_t i,ls,lt;
	ls = strlen(s);
	lt = strlen(t);

	for (i=0; i <= (ls - lt); ++i)
		if (strnicmp(s+i,t,lt) == 0)
			return s+i;

	return NULL;
}


#define LMAX	512
#define PMAX	512
#define ANY	    '?'
#define CLASS	'['
#define ENDCLASS ']'
#define ESCAPE  '\\'
#define STAR	'*'
#define RANGE '-'
#define ENDPAT  0

int uxmatch(char *line, char *pattern)

/* (partial) line to match	   */
/* (partial) pattern to match	 */
{
	register char   *l;	    /* Current line pointer	      */
	register char   *p;	    /* Current pattern pointer      */
	int		   op;	      	/* Pattern operation	      */
	char inclass = 0, lowvalue = 0, foundmatch = 0, escape = 0;
	l = line;
	p = pattern;
	if (*p == ENDPAT)
		return 1;

	while ((op = *p++) != ENDPAT) {
		escape = 0;
		if (op == '\\') {
			op = *p++;
			if (op)		
				escape = 1;
			else {	/* pattern ends with \. look for \ */
				--p;
				op = '\\';
			}
		}

		if (inclass) {
			/* currently scanning a class spec [abc-de] */
			if (!foundmatch) {
				/* and we haven't found a match yet */
				if (lowvalue) {
				/* but we are in a range c-e */
				/* op is high value of range */
					if (tolower(*l) >= lowvalue && tolower(*l) <= tolower(op)) {
						/* we found a match */
						foundmatch = 1;
						lowvalue = 0;
						continue;
					}
				/* no match found, return failure */
					lowvalue = 0;
				}
				else {
					/* not in a range */
					if (*p == RANGE)  {
						/* in a range, op is low value */
						lowvalue = (char) tolower(op);
						++p;
						continue;
					}
					else
						lowvalue = 0;

					if (tolower(*l) == tolower(op)) {
						/* found match, whoopie */
						foundmatch = 1;
						continue;
					}
				}
			}
			if (op == ENDCLASS && !escape) {
				if (!foundmatch)					
					return 0;		/* not found, sorry charlie */
				inclass = 0;
				++l;
			}
		}
		else {
			switch (op) {
			case ANY:
				if (*l++ == '\0')
					return(0);
				break;
			case CLASS:
				inclass = 1;
				lowvalue = 0;
				foundmatch = 0;
				break;				
			case STAR:		 /* Zero or more characters	*/
																																										/* match line against rest of pattern */
				if (*l == 0 && *p ==0)
					// Currently looking for *, 0 or more characters
					// The strings is empty
					// If the pattern ends with * then success; else failure
					return 1;
				while (*l)
					if (uxmatch(l++,p))
						return 1;
				return 0;		 /* Nothing else worked */
			default:
				if (tolower(*l) != tolower(op))
					return(0);
				l++;
				break;
			}
		}
	}
	return *l == 0 ? 1 : 0;   /* if everything matches and at end of string, ok */
}



void help() {	
FILE *hlp = stderr;
fprintf(hlp,"INCLUDE Version 1.2, %s\n",__DATE__);
fprintf(hlp,"\nCopyright 2003,2004 Norm London, London Computing.\nnorm_london@londoncomputing.com\n\n");
fprintf(hlp,"This program is free software; you can redistribute it and/or modify\n");
fprintf(hlp,"it under the terms of the GNU General Public License as published by\n");
fprintf(hlp,"the Free Software Foundation; either version 2 of the License, or\n");
fprintf(hlp,"(at your option) any later version.\n\n");
fprintf(hlp,"This program is distributed in the hope that it will be useful,\n");
fprintf(hlp,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
fprintf(hlp,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
fprintf(hlp,"GNU General Public License for more details.\n\n");
fprintf(hlp,"You should have received a copy of the GNU General Public License\n");
fprintf(hlp,"along with this program; if not, write to the Free Software\n");
fprintf(hlp,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");

fprintf(stderr,"Usage is: INCLUDE [{[-U] [-S]|-?}] filter.dat <infile >outfile\n");
fprintf(stderr,"Copies standard input to standard output while applying filters\n");
fprintf(stderr,"from filter.dat.  If any line in filter.dat is a sub-string of\n");
fprintf(stderr,"a line read from stdin, that line is copied to stdout.\n");
fprintf(stderr,"You may include wildcard charcters ? and * in the filters.\n");
fprintf(stderr,"? stands for any character while * stands for any one or more characters.\n");
fprintf(stderr,"Returns 0 if one or more records are included,\n");
fprintf(stderr,"        1 if all records are included,\n");
fprintf(stderr,"        2 if no records are included,\n");
fprintf(stderr,"        4 if this help is displayed or an error occurs.\n\n");
fprintf(stderr,"The -U flag indicates that case should be ignored.\n");
fprintf(stderr,"    -? flag prints this help.\n");
fprintf(stderr,"    -S removes an entire section if a filter expression\n");
fprintf(stderr,"       is found within a Windows .INI or .REG file, i.e.\n");
fprintf(stderr,"       within [ ... ]\n");
}

int main(int argc,char *argv[]) {
    /* Usage is: INCLUDE [-U] [-S] [filter] <file1 >file2 */
	/* Copies stdin to stdout, excluding those records lines containing any */
	/* of the exclusion terms stored in filter. */
	/* -U -- ignore case */
    /* -S -- Remove entire section if filter expression is found within [ ... ]*/
	struct flist {
		struct flist *next;
		char *exp;
	} 
	*ff,*cf, *mf;

	FILE *filter;

	int arg;
	size_t len;
	char ignore_case = 0, *found, *s1;
   	char delete_section = 0, skipping_section = 0;
	long inkt=0, otkt = 0;

	ff = NULL;
	cf = ff;
	filter = NULL;

	for (arg=1;arg < argc; ++arg) {
		if (stricmp(argv[arg],"?") == 0 || stricmp(argv[arg],"-?") == 0) {
			help();
			exit(4);
		}
		if (stricmp(argv[arg],"-U") == 0)
			ignore_case = 1;
        else if (stricmp(argv[arg],"-S") == 0)
            delete_section = 1;
		else {
			filter = fopen(argv[arg],"r");
			if (filter == NULL) 
				fprintf(stderr,"Unable to open: %s\n",argv[arg]);
			else {
				fprintf(stderr,"Reading filters from: %s\n",argv[arg]);
				for (s1=s+1; fgets(s1,sizeof(s)-2,filter) != NULL; s1=s+1) {
					if (s1[0] == '\n')
						continue;
					// force string to start and end with an * to force a substring match
					// anywhere in the string
					if (*s1 != '*') {
						s1 = s;
						*s1='*';
					}
					len = strlen(s1);
					// Remove trailing line feeds
					if (s1[len-1] == '\n')
						s1[--len] = 0; 
					// force pattern to end with an *
					if (*(s1+len-1) != '*')
						strcat(s1,"*");
					// Allocate a new entry in the linked list for the current pattern
					mf = (struct flist *) malloc(sizeof(struct flist));
					if (mf == NULL) {
						fprintf(stderr,"Out of Memory\n");
						exit(4);
					}
					// insert into the linked list
					if (ff == NULL)
						ff = mf;
					else
						cf->next = mf;
					// allocate memory to hold the pattern and copy into linked list
					cf = mf;
					cf->exp = (char *) malloc((len=strlen(s1))+1);
					cf->next = NULL;
					if (cf->exp == NULL) {
						fprintf(stderr,"Out of Memory\n");
						exit(4);
					}
/*					if (ignore_case) strupr(s); */
					strcpy(cf->exp,s1);
					/*	fprintf(stderr,"ff=%4x, cf=%x, cf->exp=%s, cf->next=%x\n",ff,cf,cf->exp,cf->next); */
				}
			}
		}
	}

	if (filter != NULL)
		fclose(filter);

/*	fprintf(stderr,"Excluding files containing the following:\n"); 
  	for (cf=ff;cf != NULL;cf = cf->next)
		fprintf(stderr,"ff=%4x, cf=%x, cf->exp=%s, cf->next=%x\n",
		ff,cf,cf->exp,cf->next);
*/

	/* The list of filter statments has been built, now do the copy */
	while(fgets(s,sizeof(s),stdin) != NULL) {
		++inkt;
		/* read in a line, apply filter, write out */
		/* if (ignore_case)	strupr(s); */
        if (skipping_section)
            if (s[0] == '[')
                skipping_section = 0;
            else
                continue;

		for (cf=ff;cf != NULL; cf = cf->next) {
//			if ((found = stristr(s,cf->exp)) != NULL)
			if (uxmatch(s,cf->exp)) {
				found = s;
				break;
			}
			else
				found = NULL;
		}		
		if (found != NULL) {
			++otkt;
			fputs(s,stdout);
		}
        else if ((s[0] == '[') && delete_section)
            skipping_section = 1;

	}

	cf = ff;
	while (cf != NULL) {
		ff = cf;
		cf = cf->next;
		free(ff);
	}

	fprintf(stderr,"%ld records read, %ld records written.\n",inkt,otkt);
	exit(otkt==0 ? 2: (otkt==inkt  ? 1 : 0));
}

			
				
		
